#ifndef MHACHE_ARMOR_H
#define MHACHE_ARMOR_H

#include <string>
#include <array>
#include <vector>
#include <memory>
#include <set>
#include <functional>
#include <map>

#include "mhache_skill.h"
class MHacheSkill;

template <class P, class LessOperator = std::less<P>>
struct PtrLessComparator {
    using is_transparent = void;

    bool operator () (const P* a, const P* b) const
    {
        if (!a) return true;
        if (!b) return false;
        return LessOperator()(*a, *b);
    }

    bool operator() (std::string_view a, const P* b) const
    {
        if (!b) return false;
        return LessOperator()(a, *b);
    }

    bool operator() (const P* a, std::string_view b) const
    {
        if (!a) return false;
        return LessOperator()(*a, b);
    }
};

using MappedSkillLevels =
    std::map<const MHacheSkill*, unsigned char, 
        PtrLessComparator<MHacheSkill, MHacheSkillLessOperator>>;

class MHacheArmor {
public:
    typedef enum: unsigned char {
        FIRE,
        WATER,
        THUNDER,
        ICE,
        DRAGON,
        count
    } ResistanceElement;

    using ElementalResistances = std::array<char, ResistanceElement::count>;

    class Part {
    public:
        enum Label : unsigned char {
            HEAD,
            CHEST,
            ARMS,
            WAIST,
            LEGS,
            count
        };

        static constexpr std::array<std::string_view, Label::count> label_names =
        { "head", "chest", "arms", "waist", "legs" };
    
    private:
        friend class MHacheArmor;

        Label                       _label;
        std::string                 _name;
        std::vector<unsigned char>  _slots;
        MappedSkillLevels           _skills;
        const MHacheArmor*          _armor;

        Part();

    public:
        ~Part() = default;

        Part(const Part&);
        Part(Part&&);

        Part& operator =(const Part&);
        Part& operator =(Part&&);

        bool operator  < (const Part& other) const;
        bool operator == (const Part& other) const;

        const std::string&                  name()          const;
        const std::vector<unsigned char>&   slots()         const;
        const MappedSkillLevels&            skills()        const;
        ElementalResistances                resistances()   const;

        unsigned char has_skill(std::string_view skill_name) const;

        Part& is_called(std::string_view name);
        Part& define_jewel_slots(std::vector<unsigned char>&& slots);
        Part& add_skill(const MHacheSkill* skill, unsigned char level);
        Part& detach();
    };

    using PartList  = std::array<Part, Part::Label::count>;
    using LabelList = std::vector<Part::Label>;

private:
    std::string             _name;
    unsigned char           _rarity;
    unsigned short          _defense;
    ElementalResistances    _resistances;
    PartList                _parts;
    
private:
    MHacheArmor();

public:
    MHacheArmor(std::string_view name, unsigned char rarity, unsigned short defense, 
            ElementalResistances&& resistances);
            
    MHacheArmor(const MHacheArmor&);
    MHacheArmor(MHacheArmor&&);

    ~MHacheArmor() = default;

    MHacheArmor& operator=(const MHacheArmor&);
    MHacheArmor& operator=(MHacheArmor&&);

    bool operator  < (const MHacheArmor& armor) const;
    bool operator == (const MHacheArmor& armor) const;

    const std::string&          name()          const;
    unsigned char               rarity()        const;
    unsigned char               defense()       const;
    const ElementalResistances& resistances()   const;

    LabelList attached_parts() const;
    LabelList search_for_skill(std::string_view skill_name) const;

    const Part& operator[](size_t label) const;
    Part&       operator[](size_t label);

    const Part& operator[](Part::Label label) const;
    Part&       operator[](Part::Label label);
};

template<>
struct std::hash<MHacheArmor>
{
    std::size_t operator()(const MHacheArmor& armor) const noexcept
    {
        return std::hash<std::string>{}(armor.name());
    }
};

struct MHacheArmorLessOperator {
    using is_transparent = void;

    bool operator() (const MHacheArmor& a, const MHacheArmor& b) const { return a < b; }
    bool operator() (const MHacheArmor& a, std::string_view b) const { return a.name() < b; }
    bool operator() (std::string_view a, const MHacheArmor& b) const { return a < b.name(); }
};

#endif
