#ifndef MHACHE_FACTORY_H
#define MHACHE_FACTORY_H

#include <filesystem>
#include <set>

#include <toml++/toml.h>

#include "mhache_skill.h"
#include "mhache_armor.h"

class MHacheFactory {
private:
    std::set<MHacheSkill, MHacheSkillLessOperator> _skills;
    std::set<MHacheArmor, MHacheArmorLessOperator> _armors;

    MHacheFactory() = default;

    bool parse_root_table(const std::filesystem::path&, toml::table&);

    template <typename T>
    static bool deserialize_table_node(const toml::table& table, std::string_view node_name, T& node_value)
    {
        if constexpr (std::is_integral<T>::value)
        {
            const auto node = table.get_as<int64_t>(node_name);

            if (node) {
                node_value = static_cast<T>(node->template ref<int64_t>());
                return true;
            }

            return false;
        }
        else
        {
            const auto node = table.get_as<T>(node_name);

            if (node) {
                node_value = node->template ref<T>();
                return true;
            }

            return false;
        }
    }

    template <typename T>
    static bool deserialize_table_node(const toml::table& table, std::string_view node_name, 
        std::vector<T>& listof_values)
    {
        const auto array = table.get_as<toml::array>(node_name);

        if (array) {
            listof_values.reserve(array->size());

            for (const auto& entry: *array) {
                if constexpr (std::is_integral<T>::value)
                {
                    const auto wrap = entry.as<int64_t>();

                    if (wrap)
                        listof_values.push_back(static_cast<T>(wrap->template ref<int64_t>()));
                }
                else
                {
                    const auto wrap = entry.as<T>();

                    if (wrap)
                        listof_values.push_back(wrap->template ref<T>());
                }
            }
        }

        return !(listof_values.empty());
    }

    template <typename T, size_t Length>
    static bool deserialize_table_node(const toml::table& table, std::string_view node_name, 
        std::array<T, Length>& listof_values)
    {
        const auto array = table.get_as<toml::array>(node_name);

        if (array && array->size() == Length) {
            for (size_t i = 0; i < Length; ++i) {
                if constexpr (std::is_integral<T>::value)
                {
                    const auto wrap = array->get_as<int64_t>(i);

                    if (wrap)
                        listof_values[i] = static_cast<T>(wrap->template ref<int64_t>());
                }
                else
                {
                    const auto wrap = array->get_as<T>(i);

                    if (wrap)
                        listof_values[i] = wrap->template ref<T>();
                }
            }
        }

        return !(listof_values.empty());
    }

public:
    ~MHacheFactory() = default;

    static MHacheFactory& singleton();

    bool load_skills(const std::filesystem::path&);
    bool load_armors(const std::filesystem::path&);

    inline const auto& available_skills() const { return _skills; }
    inline const auto& available_armors() const { return _armors; }

    void find_builds(const MappedSkillLevels& skill_names) const;
};

#endif