#include <utility>

#include "mhache_armor.h"

// -- PART --
MHacheArmor::Part::Part()
    : _label(MHacheArmor::Part::count)
    , _name("")
    , _slots({})
    , _skills({})
    , _armor(nullptr)
{}

MHacheArmor::Part::Part(const Part& other)
    : MHacheArmor::Part()
{
    *this = other;
}

MHacheArmor::Part::Part(Part&& other)
    : MHacheArmor::Part()
{
    *this = std::move(other);
}

MHacheArmor::Part&
MHacheArmor::Part::operator =(const Part& other)
{
    if (this != &other) {
        _label  = other._label;
        _name   = other._name;
        _slots  = other._slots;
        _skills = other._skills;
        _armor  = other._armor;
    }

    return *this;
}

MHacheArmor::Part&
MHacheArmor::Part::operator =(Part&& other)
{
    if (this != &other) {
        _label  = std::exchange(other._label, MHacheArmor::Part::count);
        _name   = std::exchange(other._name, "");
        _slots  = std::exchange(other._slots, {});
        _skills = std::exchange(other._skills, {});
        _armor  = std::exchange(other._armor, nullptr);
    }

    return *this;
}

bool
MHacheArmor::Part::operator <(const MHacheArmor::Part& other) const
{
    return _name < other._name;
}

bool
MHacheArmor::Part::operator ==(const MHacheArmor::Part& other) const
{
    return _name == other._name;
}

const std::string&
MHacheArmor::Part::name() const
{
    return _name;
}

const std::vector<unsigned char>&   
MHacheArmor::Part::slots() const
{
    return _slots;
}

const MappedSkillLevels&
MHacheArmor::Part::skills() const
{
    return _skills;
}

MHacheArmor::ElementalResistances
MHacheArmor::Part::resistances() const
{
    MHacheArmor::ElementalResistances resistances;

    if (_armor) {
        const size_t nb_parts = _armor->attached_parts().size();

        if (nb_parts > 0) {
            resistances = _armor->resistances();

            for (char& res: resistances)
                res /= nb_parts;
        }
    }

    return resistances;
}

unsigned char
MHacheArmor::Part::has_skill(std::string_view name) const
{
    unsigned char level = 0;

    const auto it = _skills.find(name);
    if (_skills.end() != it)
        level = it->second;

    return level;
}

MHacheArmor::Part&
MHacheArmor::Part::is_called(std::string_view name)
{
    _name = name;
    return *this;
}

MHacheArmor::Part&
MHacheArmor::Part::define_jewel_slots(std::vector<unsigned char>&& slots)
{
    _slots = std::exchange(slots, {});
    return *this;
}

MHacheArmor::Part&
MHacheArmor::Part::add_skill(const MHacheSkill* skill, unsigned char level)
{
    if (skill && level > 0 && level <= skill->levels().size())
        _skills[skill] = level;

    return *this;
}

MHacheArmor::Part&
MHacheArmor::Part::detach()
{
    *this = MHacheArmor::Part();
    return *this;
}
// -- PART --
// -------------------------------------------------------------------------------------------------
// -- ARMOR --
MHacheArmor::MHacheArmor()
    : _name("")
    , _rarity(0)
    , _defense(0)
    , _resistances({ 0, 0, 0, 0, 0 })
    , _parts({})
{}

MHacheArmor::MHacheArmor(std::string_view name, unsigned char rarity, unsigned short defense, 
        ElementalResistances&& elemental_resistances)
    : _name(name)
    , _rarity(rarity)
    , _defense(defense)
    , _resistances(std::move(elemental_resistances))
    , _parts({})
{}

MHacheArmor::MHacheArmor(const MHacheArmor& other)
    : MHacheArmor()
{
    *this = other;
}

MHacheArmor::MHacheArmor(MHacheArmor&& other)
    : MHacheArmor()
{
    *this = std::move(other);
}

MHacheArmor&
MHacheArmor::operator =(const MHacheArmor& other)
{
    if (this != &other) {
        _name           = other._name;
        _rarity         = other._rarity;
        _defense        = other._defense;
        _resistances    = other._resistances;
        _parts          = other._parts;
    }

    return *this;
}

MHacheArmor&
MHacheArmor::operator =(MHacheArmor&& other)
{
    if (this != &other) {
        _name           = std::exchange(other._name, "");
        _rarity         = std::exchange(other._rarity, 0);
        _defense        = std::exchange(other._defense, 0);
        _resistances    = std::exchange(other._resistances, {});
        _parts          = std::exchange(other._parts, {});
    }

    return *this;
}

bool
MHacheArmor::operator <(const MHacheArmor& other) const
{
    return _name < other._name;
}

bool
MHacheArmor::operator ==(const MHacheArmor& other) const
{
    return _name == other._name;
}

const std::string&
MHacheArmor::name() const
{
    return _name;
}

unsigned char
MHacheArmor::rarity() const
{
    return _rarity;
}

unsigned char
MHacheArmor::defense() const
{
    return _defense;
}

const MHacheArmor::ElementalResistances&
MHacheArmor::resistances() const
{
    return _resistances;
}

MHacheArmor::LabelList
MHacheArmor::attached_parts() const
{
    LabelList labels;

    for (const auto& p: _parts)
        if (p._armor)
            labels.push_back(p._label);

    return labels;
}

MHacheArmor::LabelList
MHacheArmor::search_for_skill(std::string_view skill_name) const
{
    LabelList labels;

    for (const auto& p: _parts)
        if (p.has_skill(skill_name))
            labels.push_back(p._label);

    return labels;
}

const MHacheArmor::Part&
MHacheArmor::operator[](size_t label) const
{
    return _parts[label];
}

MHacheArmor::Part&
MHacheArmor::operator[](size_t label)
{
    _parts[label]._armor = this;
    _parts[label]._label = MHacheArmor::Part::Label(label);
    return _parts[label];
}

const MHacheArmor::Part& 
MHacheArmor::operator[](MHacheArmor::Part::Label label) const
{
    return _parts[static_cast<size_t>(label)];
}

MHacheArmor::Part& 
MHacheArmor::operator[](MHacheArmor::Part::Label label)
{
    return _parts[static_cast<size_t>(label)];
}
