#include "mhache_window.h"

#include <array>
#include <csignal>
#include <iostream>
#include <filesystem>

#include "imgui.h"
#include "imgui_impl_glfw.h"
#include "imgui_impl_opengl3.h"

#include <GLFW/glfw3.h>

#include "mhache_color.h"
#include "mhache_factory.h"

volatile std::sig_atomic_t MHacheWindow::_signal_status = 0;

MHacheWindow::MHacheWindow(int width, int height)
    : _width(width), _height(height), _window(nullptr)
{
    glfwSetErrorCallback(&MHacheWindow::callback_glfw_error);

    if (GLFW_TRUE == glfwInit()) {
        glfwWindowHint(GLFW_CONTEXT_VERSION_MAJOR,  _opengl_major_version);
        glfwWindowHint(GLFW_CONTEXT_VERSION_MINOR,  _opengl_minor_version);
        glfwWindowHint(GLFW_SCALE_TO_MONITOR,       GLFW_TRUE);
        glfwWindowHint(GLFW_RESIZABLE,              GLFW_FALSE);
        glfwWindowHint(GLFW_DECORATED,              GLFW_TRUE);

        _window = glfwCreateWindow(_width, _height, _title, nullptr, nullptr);

        if (_window) {
            std::signal(SIGINT,  &MHacheWindow::callback_signals);
            std::signal(SIGTERM, &MHacheWindow::callback_signals);
            std::signal(SIGABRT, &MHacheWindow::callback_signals);

            glfwSetKeyCallback(_window, &MHacheWindow::callback_glfw_keyboard);

            glfwMakeContextCurrent(_window);
            glfwSwapInterval(1);

            ImGui::CreateContext();
            ImGui_ImplGlfw_InitForOpenGL(_window, true);
            ImGui_ImplOpenGL3_Init(_GLSL_version);

            const std::filesystem::path filefont = "../../fonts/Sweet16mono.ttf";
            
            if (std::filesystem::exists(filefont)) {
                ImGui::GetIO().Fonts->AddFontFromFileTTF(filefont.u8string().c_str(), 16);
            }
                
            else
                ImGui::GetIO().Fonts->AddFontDefault();
        }
    }

    MHacheFactory& mhf = MHacheFactory::singleton();
    mhf.load_skills("../../skills.toml");
    mhf.load_armors("../../armors.toml");
}

MHacheWindow::~MHacheWindow()
{
    ImGui_ImplOpenGL3_Shutdown();
    ImGui_ImplGlfw_Shutdown();
    ImGui::DestroyContext();

    glfwDestroyWindow(_window);
    glfwTerminate();
}

void
MHacheWindow::callback_signals(int signal)
{
    _signal_status = signal;
}

void
MHacheWindow::callback_glfw_error(int error, const char* description)
{
    std::cerr << "GLFW error (" << error << ") = " << description << std::endl;
}

void
MHacheWindow::callback_glfw_keyboard(GLFWwindow* window, int key, int scancode, int action, int mods)
{
    if (GLFW_PRESS == action) 
    {
        if (GLFW_KEY_F4 == key && (GLFW_KEY_LEFT_ALT == key || GLFW_KEY_RIGHT_ALT == key)) { 
            glfwSetWindowShouldClose(window, GLFW_TRUE);
        }  
    }
}

void
MHacheWindow::new_frame()
{
    ImGui_ImplOpenGL3_NewFrame();
    ImGui_ImplGlfw_NewFrame();
    ImGui::NewFrame();
}

void
MHacheWindow::render_frame()
{
    ImGui::Render();
    glfwGetFramebufferSize(_window, &_width, &_height);

    glViewport(0, 0, _width, _height);
    glClearColor(0.0f, 0.0f, 0.0f, 1.0f);
    glClear(GL_COLOR_BUFFER_BIT);
    ImGui_ImplOpenGL3_RenderDrawData(ImGui::GetDrawData());

    glfwSwapBuffers(_window);
}

void
MHacheWindow::build_widgets()
{
    ImGuiWindowFlags window_flags = 0;
    window_flags |= ImGuiWindowFlags_MenuBar;
    window_flags |= ImGuiWindowFlags_NoSavedSettings;
    window_flags |= ImGuiWindowFlags_NoMove;
    window_flags |= ImGuiWindowFlags_NoResize;
    window_flags |= ImGuiWindowFlags_NoCollapse;
    window_flags |= ImGuiWindowFlags_NoTitleBar;

    const ImGuiViewport* main_viewport = ImGui::GetMainViewport();
    if (main_viewport) {
        ImGui::SetNextWindowPos(main_viewport->WorkPos);
        ImGui::SetNextWindowSize(main_viewport->WorkSize);
        ImGui::SetNextWindowSizeConstraints(main_viewport->WorkSize, main_viewport->WorkSize);
    }
    
    ImGui::Begin(_title, nullptr, window_flags);
    //--
    build_menu();

    static std::vector<const MHacheSkill*> filtered_skills;
    static ImGuiTextFilter filter;

    if (ImGui::CollapsingHeader("Skills")) {
        filtered_skills.clear();
        
        filter.Draw();

        const auto& skills = MHacheFactory::singleton().available_skills();

        for (const auto& skill: skills) {
            if (filter.PassFilter(skill.name().c_str())) {
                const MHacheColor* color = skill.color();
                if (!color) {
                    std::cerr << "Missing color for " << skill.name() << std::endl;
                    continue;
                }

                const auto rgba = color->channels<float>();

                ImGui::TextColored(
                    ImVec4(rgba[0], rgba[1], rgba[2], rgba[3]),
                    "%s", skill.name().c_str()
                );

                if (ImGui::IsItemHovered()) {
                    ImGui::BeginTooltip();
                    ImGui::PushTextWrapPos(ImGui::GetFontSize() * 35.0f);
                    ImGui::TextUnformatted(skill.description().c_str());
                    ImGui::PopTextWrapPos();
                    ImGui::EndTooltip();
                }

                filtered_skills.push_back(&skill);
            }
        }       
    }

    if (ImGui::CollapsingHeader("Armors")) {
        const auto& armors = MHacheFactory::singleton().available_armors();
    
        for (const auto& armor: armors) {
            for (const auto skill: filtered_skills) {
                if (!armor.search_for_skill(skill->name()).empty()) {
                    ImGui::Text("%s", armor.name().c_str());
                    break;
                }
            }
        }

        MappedSkillLevels seeklevels;
        for (const auto skill: filtered_skills) {
            seeklevels.insert({ skill, 1 });
        }

        MHacheFactory::singleton().find_builds(seeklevels);
    }

    //--
    ImGui::End();
}

void
MHacheWindow::build_menu()
{
    if (ImGui::BeginMenuBar()) {
        if (ImGui::BeginMenu("App")) {

            if (ImGui::BeginMenu("Resize")) {
                if (ImGui::MenuItem("640x480"))     { resize(640,   480); }
                if (ImGui::MenuItem("800x600"))     { resize(800,   600); }
                if (ImGui::MenuItem("1024x768"))    { resize(1024,  768); }
                if (ImGui::MenuItem("1280x720"))    { resize(1280,  720); }
                ImGui::EndMenu();
            }

            if (ImGui::MenuItem("Quit", "Alt+F4")) { 
                glfwSetWindowShouldClose(_window, GLFW_TRUE); 
            }

            ImGui::EndMenu();
        }
        ImGui::EndMenuBar();
    }
}

void
MHacheWindow::resize(int width, int height)
{
    _width  = width;
    _height = height;

    if (!ready()) return;

    GLFWmonitor* monitor = glfwGetPrimaryMonitor();
    if (!monitor) return;

    const GLFWvidmode* videomode = glfwGetVideoMode(monitor);
    if (!videomode) return;

    glfwSetWindowPos(_window, (videomode->width - _width) * 0.5, (videomode->height - _height) * 0.5);
    glfwSetWindowSize(_window, _width, _height);
}

bool
MHacheWindow::ready()
{
    return SIGINT  != _signal_status
        && SIGTERM != _signal_status
        && SIGABRT != _signal_status
        && _window 
        && !glfwWindowShouldClose(_window);
}

void
MHacheWindow::display()
{
    while (ready()) {
        glfwPollEvents();
        new_frame();
        build_widgets();
        render_frame();
    }
}
